/*
**
**   This code is a part of the ETSI Technical Specification 06.77,
**   Digital cellular telecommunications system (Phase 2+);
**   Minimum Performance Requirements for Noise Suppresser Application to the AMR Speech Encoder
**
*/

/*
       =============================================================
       COPYRIGHT NOTE: This source code, and all of its derivations,
       is subject to the "ITU-T SOFTWARE TOOLS' GENERAL PUBLIC LICENSE". 
	   Please have it  read  in    the  distribution  disk,   or  in  
	   the  ITU-T Recommendation G.191 on "SOFTWARE TOOLS FOR SPEECH AND  
	   AUDIO CODING STANDARDS".
       =============================================================
*/


/*
**
**   OBJECTIVE MEASURES FOR CHARACTERISATION OF AMR/NS SOLUTIONS
**
**
**   MODULE:         process.c, functions for the processing of the speech database samples
**                     by the AMR speech codec and AMR/NS solution and the computation of the
**                     objective measures for the characterisation of NS performance
**   ORIGINAL BY:    
**      Beghdad Ayad     Nokia Research Center    Finland
**      Erkki Paajanen   Nokia Research Center    Finland
**   DATE:           19/Jan/2001
**   RELEASE:        1.00
**   FUNCTIONS:      Process ........... processing of input speech file by AMR and AMR/NS,
**                                       calling of ObjMeasRes for objective measures,
**                                       routing objective measure results to calling function
**                   ObjMeasRes ........ computation of objective SNRI and NPLR measures for
**                                       input speech sample using clean speech reference,
**                                       AMR speech codec output and AMR/NS solution output
**   PROTOTYPES:     in main.h
**
**
**   MODIFICATION HISTORY:
**      15/Jun/00 v0.90 Release for inclusion in ETSI TS 06.77.   <erkki.paajanen@nokia.com>
**      19/Jan/01 v1.00 CR release: Inclusion of "ITU General Public License" terms.
**                                                                <erkki.paajanen@nokia.com>
**
**
*/


/*
** Copyright (c) 1999-2001 Nokia Corporation
*/


/* ... Includes in general ... */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "main.h"

/* ... Include of speech-voltmeter-related routines ... */
#include "sv-p56.h"

/* ... Include of utilities ... */
#include "ugst-utl.h" 


#define Max(a,b) (a > b ? a : b)


/*-----------------------------------------------------------------------*
 *                          constants
 *-----------------------------------------------------------------------*/

#define FRAME_SIZE     80

#define TH_H           -1
#define TH_M           -10
#define TH_L           -16
#define TH_Nh          -19
#define TH_Nl          -34


#define ksi            0.00001f


/*---------------------------------------------------------------------------
 *                                Functions
 *--------------------------------------------------------------------------*/
static Report *ObjMeasRes(char *CleanFilename, char *AmrFilename, char *AmrNsFilename);



/*---------------------------------------------------------------------------
 
        Report *Process(int run_mode, char *CleanFilename, char *NoisyFilename,
                ======= char *AmrFilename, char *AmrNsFilename, char *AmrMode)

        Description:
        ~~~~~~~~~~~~

        First checks existence of files claimed in DescDb.c, then according to
        control argument 'run_mode', either
         - (1) produces dummy report
         - (2) assuming processed files to exist, calls ObjMeasRes for the computation
           of the objective NS measures
         - (3) processes AMR and AMR/NS output in temporary files and calls ObjMeasRes
         - (4) processes and stores AMR and AMR/NS output and calls ObjMeasRes

        Arguments:
        ~~~~~~~~~~
        Name:        I/0: Type:  Use:
        run_mode      I   int    Mode of operation
        CleanFilename I   char * File name for clean speech reference
        NoisyFilename I   char * File name for noisy speech
        AmrFilename   I   char * File name for AMR speech codec output
        AmrNsFilename I   char * File name for AMR/NS solution output
        AmrMode       I   char * AMR speech codec mode to be used.

        Return value:
        ~~~~~~~~~~~~~
        Returns the data structure (Report) comprising the necessary data for the
        computed objective NS measures.

        Functions used:        ObjMeasRes
        ~~~~~~~~~~~~~~~

        Prototype:             in main.h
        ~~~~~~~~~~

        Author:                Beghdad Ayad     Nokia Research Center   Finland
        ~~~~~~~                Erkki Paajanen   Nokia Research Center   Finland
                               <erkki.paajanen@nokia.com>

        Modification history:
        ~~~~~~~~~~~~~~~~~~~~~
        Date         Version   Description
        07/Jun/00    0.10      Preliminary version for review in SMG11
                               <erkki.paajanen@nokia.com>
        19/Jan/00    1.00      CR release for TS 06.77.
                               <erkki.paajanen@nokia.com>

 *--------------------------------------------------------------------------*/
/*
** Copyright (c) 1999-2001 Nokia Corporation
*/
Report *Process(int run_mode, char *CleanFilename, char *NoisyFilename,
                char *AmrFilename, char *AmrNsFilename, char *AmrMode)
{
   static Report report;
   FILE *f;
   char commandLine[500];
   Report *resout;


/*
 * check that file exists
 */
   f = fopen(CleanFilename, "rb");
   if(!f) {
      fprintf(stderr, "%s file doesn't exists\n", CleanFilename);
      exit(1);
   }
   fclose(f);

   f = fopen(NoisyFilename, "rb");
   if(!f) {
      fprintf(stderr, "%s file doesn't exists\n", NoisyFilename);
      exit(1);
   }
   fclose(f);


   if (run_mode == 1)   /* For file search test */
   {

      /* return dummy report */
      report.NbFr     = 500 + 300 + 200 + 100 + (345) ;
      report.Ksph     = 500;
      report.Kspm     = 300;
      report.Kspl     = 200;
      report.Knse     = 100;

      report.SNRI_h   = 20;
      report.SNRI_m   = 10;
      report.SNRI_l   = 5;

      report.NPLR     = -20;

      report.Ktot     = 500 + 300 + 200;
      report.SNRI     = (500*20+300*10+200*5)/(1000);

      return(&report);

   }

   else /* run_mode == 2...4 */
   {
      
      if (run_mode == 3 || run_mode == 4) /* Conduct speech coding and NS processing */
      {

         strcpy(commandLine, COMMAND_LINE_AMR);
         strcat(commandLine, " ");
         strcat(commandLine, AmrMode);
         strcat(commandLine, " ");
         strcat(commandLine, NoisyFilename);
         strcat(commandLine, " bit.tmp");
         system(commandLine);

         strcpy(commandLine, COMMAND_LINE_DECODER);
         strcat(commandLine, " bit.tmp ");
         strcat(commandLine, AmrFilename);
         system(commandLine);

         strcpy(commandLine, COMMAND_LINE_AMRNS);
         strcat(commandLine, " ");
         strcat(commandLine, AmrMode);
         strcat(commandLine, " ");
         strcat(commandLine, NoisyFilename);
         strcat(commandLine, " bit.tmp");
         system(commandLine);

         strcpy(commandLine, COMMAND_LINE_DECODER);
         strcat(commandLine, " bit.tmp ");
         strcat(commandLine, AmrNsFilename);
         system(commandLine);

      } /* if (run_mode == 3 || run_mode == 4) */
         
     /*
            Results of the ObjMeas.
     */

      resout = ObjMeasRes(CleanFilename, AmrFilename, AmrNsFilename);

      /*
            Delete temporary files
      */

      if (run_mode == 3 || run_mode == 4)
      {

         strcpy(commandLine, CMD_FILE_REMOVE);
         strcat(commandLine, "bit.tmp");
         system(commandLine);

         strcpy(commandLine, CMD_FILE_REMOVE);
         strcat(commandLine, AmrFilename);
         system(commandLine);

         strcpy(commandLine, CMD_FILE_REMOVE);
         strcat(commandLine, AmrNsFilename);
         system(commandLine);

      } /* if (run_mode == 3 || run_mode == 4) */

     /*
           fill in report
     */

      report.SNRI_h   = resout->SNRI_h;
      report.SNRI_m   = resout->SNRI_m;
      report.SNRI_l   = resout->SNRI_l;
  
      report.SNRI     = resout->SNRI;
      report.NPLR     = resout->NPLR;
   
      report.NbFr     = resout->NbFr;

      report.Ksph     = resout->Ksph;
      report.Kspm     = resout->Kspm;
      report.Kspl     = resout->Kspl;
      report.Knse     = resout->Knse;

      report.Ktot     = resout->Ktot;
  

      return(&report);

   } /* if (run_mode == 1) */
}



/*---------------------------------------------------------------------------
 
        Report *ObjMeasRes(char *CleanFilename, char *AmrFilename, char *AmrNsFilename)
                ==========

        Description:
        ~~~~~~~~~~~~

        First determines the active speech level in the clean speech reference
        using the ITU-T P.56 Speech Voltmeter. Next classifies the frames in the
        clean speech reference into speech of high, medium and low power, noise
        in vicinity of speech and those belonging to none of these categories.
        Finally, computes two objective measures for the characterisation
        of noise suppresser performance in the given noisy speech file through a
        comparison of the AMR/NS output on the noisy speech to the AMR output on the
        same noisy speech using the above frame classification. The two measures are:
        1. SNRI, signal-to-noise ratio (SNR) improvement, representing the enhancement
           of the ratio of the average power in the speech frames to the average
           power in the noise frames. This measure is computed both separately in the
           high, medium and low power speech classes (SNRI_h, SNRI_m, SNRI_l,
           respectively) and as an overall measure (SNRI).
        2. NPLR, noise power level reduction, representing the reduction of the
           average power in the frames classified as noise in the vicinity of speech.

        Arguments:
        ~~~~~~~~~~
        Name:        I/0: Type:  Use:
        CleanFilename I   char * File name for clean speech reference
        AmrFilename   I   char * File name for AMR speech codec output
        AmrNsFilename I   char * File name for AMR/NS solution output

        Return value:
        ~~~~~~~~~~~~~
        Returns the data structure (Report) comprising the necessary data for the
        computed objective NS measures.

        Functions used:        init_speech_voltmeter (P.56), speech_voltmeter (P.56)
        ~~~~~~~~~~~~~~~

        Prototype:             in process.c (this file)
        ~~~~~~~~~~

        Author:                Beghdad Ayad     Nokia Research Center   Finland
        ~~~~~~~                Erkki Paajanen   Nokia Research Center   Finland
                               <erkki.paajanen@nokia.com>

        Modification history:
        ~~~~~~~~~~~~~~~~~~~~~
        Date         Version   Description
        07/Jun/00    0.10      Preliminary version for review in SMG11
                               <erkki.paajanen@nokia.com>
        19/Jan/00    1.00      CR release for TS 06.77.
                               <erkki.paajanen@nokia.com>

 *--------------------------------------------------------------------------*/
/*
** Copyright (c) 1999-2001 Nokia Corporation
*/
Report *ObjMeasRes(char *CleanFilename, char *AmrFilename, char *AmrNsFilename)
{
   static Report resout;
   long   N=FRAME_SIZE, N2 = 0;
   long   bitno = 16;
   double ActiveLeveldB;

   /* Intermediate storage variables for speech voltmeter */
   SVP56_state     state;

   FILE *CleanSpeech;
   FILE *AmrOutput;
   FILE *AmrNsOutput;


   double sf     = 8000;    /* Sampling Frequency Hz */

   int Ksph      = 0;
   int Kspm      = 0;
   int Kspl      = 0;
   int Knse      = 0;

   
   long  i,L;
   int   j;
   
   double snr_in, snr_out;
   double sum_sdB, sum_s, sum_y, sum_c;
   double sum_sphy=ksi, sum_spmy=ksi, sum_sply=ksi, sum_nsey=ksi, sum_spny=ksi;
   double sum_sphc=ksi, sum_spmc=ksi, sum_splc=ksi, sum_nsec=ksi, sum_spnc=ksi;

   float SNRI_h  ;
   float SNRI_m  ;
   float SNRI_l  ;

   float NPLR    ;

   short buffer_clean[FRAME_SIZE];
   float Buf_clean[FRAME_SIZE];

   short buffer_amr[FRAME_SIZE];
   float Buf_amr[FRAME_SIZE];

   short buffer_amrns[FRAME_SIZE];
   float Buf_amrns[FRAME_SIZE];
   

   /* Reset variables for speech level measurements */
   init_speech_voltmeter(&state, sf);

   /* open Clean Speech File  */                         
   CleanSpeech = fopen(CleanFilename, "rb");
   if(!CleanSpeech) {
      fprintf(stderr, "unable to open clean speech file %s\n", CleanFilename);
      exit(1);
   }
   
   /* open processed File (only AMR)  */                         
   AmrOutput = fopen(AmrFilename, "rb");
   if(!AmrOutput) {
      fprintf(stderr, "unable to open processed noisy speech (AMRonly) file %s\n", AmrFilename);
      exit(1);
   }

   /* open processed File (AMRNS)  */                         
   AmrNsOutput = fopen(AmrNsFilename, "rb");
   if(!AmrNsOutput) {
      fprintf(stderr, "unable to open processed noisy speech (AMRNS) file %s\n", AmrNsFilename);
      exit(1);
   }
   
   while ((L = fread(buffer_clean, sizeof(short), N, CleanSpeech)) > 0) N2++;
   rewind(CleanSpeech);


   for (i = 0; i < N2; i++)
   {
      /* Compute the ActiveLeveldB for the current CleanSpeech file */
      if ((L = fread(buffer_clean, sizeof(short), N, CleanSpeech)) > 0)
      {   
         /* ... Convert samples to float */
         sh2fl((long) L, buffer_clean, Buf_clean, bitno, 1);

         /* ... Get the active level ... */
         ActiveLeveldB = speech_voltmeter(Buf_clean, (long) L, &state);
      }
      else
      {
         fprintf(stderr, "unable to open clean speech file %s\n", CleanFilename);
         exit(1);
      }

   } /* for (i=0... */

   rewind(CleanSpeech);
   fseek(AmrOutput, (long) AMR_DELAY, SEEK_SET);
   fseek(AmrNsOutput, (long) AMRNS_DELAY, SEEK_SET);


   for (i = 0; i < N2-2; i++)
   { 
      /* ... Convert samples to float ... */
      /* read a frame from clean file */
      if((L = fread(buffer_clean, sizeof(short), N, CleanSpeech)) <= 0) {
         fprintf(stderr, "error while reading clean speech file %s\n", CleanFilename);
         exit(1);
      }
      sh2fl((long) L, buffer_clean, Buf_clean, bitno, 1);

      /* read a frame from processed file (only AMR) */
      if(fread(buffer_amr, sizeof(short), N, AmrOutput) !=(long) L) {
         fprintf(stderr, "error while reading processed noisy speech (AMR only) file %s\n", AmrFilename);
         exit(1);
      }
      sh2fl((long) L, buffer_amr, Buf_amr, bitno, 1);

      /* read a frame from processed file (AMR/NS) */
      if(fread(buffer_amrns, sizeof(short), N, AmrNsOutput) !=(long) L) {
         fprintf(stderr, "error while reading processed noisy speech (AMR/NS) file %s\n", AmrNsFilename);
         exit(1);
      }
      sh2fl((long) L, buffer_amrns, Buf_amrns, bitno, 1);

      /* clean speech power */
      sum_s = 0;
      for(j=0;j<L;j++)
      sum_s     += (Buf_clean[j]*Buf_clean[j]);

      if(sum_s<ksi)    sum_sdB = -113;
      else             sum_sdB = 10.f*log10(sum_s/N);


      /* AMR output power */
      sum_c = 0;
      for(j=0;j<L;j++) 
         sum_c    += Buf_amr[j]*Buf_amr[j];
        
    
      /* AMR+NS output power */
      sum_y = 0;
      for(j=0;j<L;j++)        
         sum_y    += Buf_amrns[j]*Buf_amrns[j];
        

      /*     ...           Get the Objective Measures        ....    */
 
      if (sum_sdB >   ActiveLeveldB + TH_H)  /* High power frames */
      { 
         sum_sphy   += (sum_y);
         sum_sphc   += (sum_c);
         Ksph++;
      }
      else if(sum_sdB>ActiveLeveldB + TH_M)  /* Medium power frames */
      { 
         sum_spmy   += (sum_y);
         sum_spmc   += (sum_c); 
         Kspm++;
      }
      else if(sum_sdB>ActiveLeveldB + TH_L)  /* Low power frames */
      { 
         sum_sply   += (sum_y); 
         sum_splc   += (sum_c); 
         Kspl++;
      }
                         
      else if((sum_sdB>ActiveLeveldB+TH_Nl)&&(sum_sdB<ActiveLeveldB+TH_Nh)) /* Only noise  frames */
      {
         sum_nsey   += (sum_y); 
         sum_nsec   += (sum_c); 
         Knse++;
      }
        
   }   /* end of the main loop i */
        

   snr_out  = (sum_sphy/Ksph)/(sum_nsey/Knse) - 1.f ;
   snr_in   = (sum_sphc/Ksph)/(sum_nsec/Knse) - 1.f ;
   if((snr_out<=ksi)||(snr_in<=ksi)) snr_out = snr_in = ksi;
      SNRI_h  = 10*(float)(log10(Max(ksi,snr_out))-log10(Max(ksi,snr_in)));


   snr_out  = (sum_spmy/Kspm)/(sum_nsey/Knse) - 1.f;
   snr_in   = (sum_spmc/Kspm)/(sum_nsec/Knse) - 1.f;
   if((snr_out<=ksi)||(snr_in<=ksi)) snr_out = snr_in = ksi;
      SNRI_m  = 10*(float)(log10(Max(ksi,snr_out))-log10(Max(ksi,snr_in)));


   snr_out = (sum_sply/Kspl)/(sum_nsey/Knse) - 1.f;
   snr_in  = (sum_splc/Kspl)/(sum_nsec/Knse) - 1.f;
   if((snr_out<=ksi)||(snr_in<=ksi)) snr_out = snr_in = ksi;
      SNRI_l  = 10*(float)(log10(Max(ksi,snr_out))-log10(Max(ksi,snr_in)));


   snr_in  = (sum_nsey)/(sum_nsec);
   NPLR    = 10*(float)(log10(Max(ksi,snr_in)));


   /**************** close file fill stat ******************/

   fclose(CleanSpeech);
   fclose(AmrOutput);
   fclose(AmrNsOutput);
   
   resout.NbFr     = N2;   /* total number of frames in current file */

   resout.Ksph     = Ksph;
   resout.Kspm     = Kspm;
   resout.Kspl     = Kspl;
   resout.Knse     = Knse;


   resout.SNRI_h   = SNRI_h;
   resout.SNRI_m   = SNRI_m;
   resout.SNRI_l   = SNRI_l;
   
   resout.NPLR     = NPLR;

   resout.Ktot     = Ksph + Kspm + Kspl;
   resout.SNRI     = (Ksph*SNRI_h + Kspm*SNRI_m + Kspl*SNRI_l)/(Ksph + Kspm + Kspl) ;


   return(&resout);
}
